/*
 * Programmer:  Timothy E. Hoff
 * File Name:   main.c
 * Written:     July 7, 1996
 * Modified:
 * Description: This module controls the program execution of sfw (Simple
 *              File Wrapper).
 *
 * Copyright (c) 1996, Timothy E. Hoff
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include "sfw.h"
#include "config.h"
#include "acl.h"
#include "sizes.h"
#include "pathnames.h"
#include "log.h"
#include "mailer.h"
#include "error.h"
#include "util.h"

#ifndef _LINT
static char rcsid[] = "@(#) $Id: main.c,v 2.3 1996/07/22 20:42:34 t07de7 Exp t07de7 $";
#endif

/*
 * Program Flow:
 * 1.  Lookup command alias (argv[0])
 * 2.  Get ACE for current user
 * 3.  Verify access privileges
 * 4.  Setup new environment
 *     a.  setuid()
 *     b.  setgid() initgroups
 *     c.  set TERM, HOME, USER, LOGNAME, etc.
 * 5.  Execute target command
 *
 */

char *cleanenv[1];


/*
 * Function Declarations
 */

void access_denied(struct cmd_alias *, char *, char **);
void access_granted(struct cmd_alias *, char *, char **);

int main(int argc, char *argv[], char *envp[])
{
    struct cmd_alias *ca;           /* cmd alias record */
    struct ace *ace;                /* access control entry */
    struct passwd *pw;
    struct group *grp;
    uid_t uid;                      /* user ID */
    char login[MAXLOGIN];           /* user name */
    char term[BUFSIZE];             /* terminal type */
    char timezone[BUFSIZE];         /* TZ setting */
    char msg[BUFSIZE];
    int cnt;                        /* open file count */
    char **argptr;                  /* pointer to new cmd line */
    char **ptr;                     /* temp pointer */
    extern char **environ;          /* current environment */
    extern char *cleanenv[];        /* new environment */
    int argcnt = 0;                 /* arguments in new cmd line */
    int i;

    tzset();                        /* setup local time zone */
    
    if ((ca = getcmdalias(strip_path(argv[0]))) == NULL) {
        fprintf(stderr, "ERROR:  sfw:  %s is not a defined resource.\n", 
                strip_path(argv[0]));
        fprintf(stderr, "Contact your system administrator.\n");
        exit(EXIT_FAILURE);
    }

    if ((pw = getpwuid(getuid())) == NULL) {
        sprintf(msg, "Could not find passwd entry for uid %d.", getuid());
        log_error(EPASSWD, msg);
        fprintf(stderr, "ERROR:  sfw:  Could not read from password file.\n");
        exit(EXIT_FAILURE);
    }

    /*
     * ACL special cases:
     * The ACL may be a '+' (plus sign) in which case all users
     * are granted access or a '-' (minus sign) in which case
     * no one receives access.
     */

    strncpy(login, pw->pw_name, MAXLOGIN);

    if (ca->acl_path[0] == '-') {
        access_denied(ca, login, argv);
        exit(EXIT_FAILURE);
    } else if (ca->acl_path[0] == '+' || getuid() == 0) {
        ;           /* access controls disabled */
    } else {
        if ((ace = getace(login, ca->acl_path, ca->acl_admin)) == NULL) {
            access_denied(ca, login, argv);
            exit(EXIT_FAILURE);
        }
    
        if (verify_user(ace, ca->tty_list, ca->req_passwd) != 0) {
            access_denied(ca, login, argv);
            exit(EXIT_FAILURE);
        }
    }

    if (*ca->new_user != '\0') {
        if ((pw = getpwnam(ca->new_user)) == NULL) {
            sprintf(msg, "User %s for cmd alias %s not found.",
                    ca->new_user, ca->alias);
            log_error(EPASSWD, msg);
            fprintf(stderr, "ERROR:  sfw:  user %s is not known.\n", 
                    ca->new_user);
            exit(EXIT_FAILURE);
        }
    } else {
       sprintf(msg, "Required attribute [user] missing from %s stanza.",
               ca->alias);
       log_error(ESYNTAX, msg);
       fprintf(stderr, "ERROR:  sfw:  Required field is missing.\n");
       fprintf(stderr, "Contact your system administrator.\n");
       exit(EXIT_FAILURE);
    }

    /*
     * If a MD5 signature is registered use it to ensure that
     * the target has not been compromised.
     */

    if (ca->md5_signature[0] != '\0') {
        if (md5_check(ca->exec_path, ca->md5_signature) != 0) {
            sprintf(msg, "%s is different than its signature.", ca->exec_path);
            log_error(ESIGN, msg);
            fprintf(stderr, "Cannot execute %s.\n", ca->exec_path);
            fprintf(stderr, "Contact your system administrator immediately.\n");
            exit(EXIT_FAILURE);
        }
    } /* no signature to match */

    /*
     * User has necessary permissions to access resource
     */

    setuid(0);          /* change real uid to superuser */

    if (ca->new_group[0] != '\0') {
        if ((grp = getgrnam(ca->new_group)) != NULL) {
            setgid(grp->gr_gid);
            initgroups(login, pw->pw_gid);
        } else {
            sprintf(msg, "Group %s for cmd alias %s not found.",
                    ca->new_group);
            log_error(EGROUP, msg);
            exit(EXIT_FAILURE);
        }
    } else {
        setgid(pw->pw_gid);
        initgroups(login, pw->pw_gid);
    }

    
    /*
     * Clean up the current environment for the target program
     */

    for (cnt = getdtablesize(); cnt > 2; cnt--) {
        close(cnt);
    }

    strncpy(term, getenv("TERM"), sizeof(term) -1);
    strncpy(timezone, getenv("TZ"), sizeof(timezone) -1);
    
    environ = cleanenv;
    cleanenv[0] = NULL;

    (void) setenv("HOME", pw->pw_dir, 1);
    (void) setenv("SHELL", pw->pw_shell, 1);
    (void) setenv("TERM", term, 0);
    (void) setenv("TZ", timezone, 0);
    (void) setenv("LOGNAME", pw->pw_name, 1);
    (void) setenv("USER", pw->pw_name, 1);
    (void) setenv("PATH", _PATH_DEFPATH, 1);

    /*
     * Save the new user ID but don't change to it yet.
     */

    uid = pw->pw_uid;

    endpwent();         /* close the password database */
    endgrent();         /* close the group database */

    /*
     * Load additional environmental variables if defined in config
     */

    if (ca->environ_path[0] != '\0') {
        (void) load_environ(ca->environ_path);
    }

    argcnt = 1 + (ca->pass_args ? argc - 1 : 0);
    for (i=0; i<MAXARGC; i++) {
        if (ca->argv[i] != NULL)
            argcnt++;
        else
            break;
    }

    if ((argptr = calloc((argcnt + 1), sizeof(char *))) == NULL) {
        fprintf(stderr, "Insufficient memory.\n");
        return -1;
    }

    *argptr = ca->exec_path;
    for (i=0, ptr=argptr; i<MAXARGC; i++) {
        if (ca->argv[i] != NULL)
            *(++ptr) = ca->argv[i];
        else
            break;
    }

    if (ca->pass_args) {
        for (i=1; i<argc; i++)
            *(++ptr) = argv[i];
    }

    access_granted(ca, login, argptr);
    if (setuid(uid) != 0) {
        sprintf(msg, "Could not set user ID to  %d.", uid);
        log_error(errno, msg);
        return -1;
    } else {
        execve(ca->exec_path, argptr, environ);
        fprintf(stderr, "Could not execute %s (E %d).", ca->exec_path, errno);
        exit(EXIT_FAILURE);
    }
    /* Not reached */
}


void access_denied(struct cmd_alias *ca, char *login, char **argv)
{
    /*
     * Record a failed access attempt.
     */
     
    log_deny(ca->log_path, login, argv);

    if (ca->auth_notify[0] != '\0') {
        sendmsg(ca->auth_notify, login, '-', argv);
    } else if (ca->admin_notify[0] != '\0') {
        sendmsg(ca->admin_notify, login, '-', argv);
    }
    
    fprintf(stderr, "Permission denied.\n");

    return;
}


void access_granted(struct cmd_alias *ca, char *login, char **argv)
{
    /*
     * Record a successful access attempt.
     */
     
    log_grant(ca->log_path, login, argv);

    if (ca->admin_notify[0] != '\0') {
        sendmsg(ca->admin_notify, login, '+', argv);
    }

    return;
}
